/*!
*	@file cook.c main source code for Cooking Assistant for Automatic Temperature Control
*   @author Adam Mendrela, Andrew Kerns
*   Email: {aem228,ajk245}@cornell.edu
*   Copyright (c) 2012, A.E. Mendrela, A.J. Kerns.
*   Available for non-commercial use, without permission, by including reference to these authors
*	Final Project, ECE4760, School of Electrical and Computer Engineering, Cornell University
*   Advisor: Dr. Bruce Land
*
*
* PINS:
*	- PORTA:	PORTA(0:1) are the analog inputs for temperature sensing
*   			PORTA(3:7) are selectable resistors
*	- PORTB:	PORTB(4:7) are NC during startup to avoid interference with ISP programming
				PORTB(0:7) are for lab keypad
*	- PORTC:	PORTC(0:7) are for Nerdkits LCD
*   - PORTD:	PORTD(0:1) are for UART
*				PORTD(4:7) are for stepper motor driving
*
*
*	A0: Ain - NTC circuit
*	A1: Ain - AD22100 circuit
*	A2: NC
*	A3: RES1
*	A4: RES2
*	A5: RES3
*	A6: RES4
*	A7: RES5
*
*	B0: Keypad
*	B1: Keypad
*	B2: Keypad
*	B3: Keypad
*	B4: Keypad
*	B5: Keypad
*	B6: Keypad
*	B7: Keypad
*
*
*	C0: LCD
*	C1: LCD
*	C2: LCD
*	C3: LCD
*	C4: LCD
*	C5: LCD
*	C6: LCD
*	C7: LCD
*
*	D0: RX
*	D1: TX
*	D2: NC
*	D3: NC
*	D4: STEPPER_1a
*	D5: STEPPER_1b
*	D6: STEPPER_2a
*	D7: STEPPER_2b
*
*/

/* - - - - PRE-PROCESSOR DIRECTIVES - - - - */

#define DEBUG
#include "cook.h"
#include <avr/eeprom.h>

#define TOGGLE_BUZZER_PIN (PORTD ^= 1<<2)

/* - - - - GLOBAL VARIABLE DEFINITIONS - - - - */

uint16_t timer_ms = 50000; //don't want it to start at zero 

uint8_t readTempCount; //count averaging

uint16_t waitMain;
uint8_t waitKeypad, waitMeas;
uint8_t mainToggle;

/* // original 
float ffactor = 0.95;
float gain_kp = 15;
float gain_ki = 0.15;
float gain_kd = 15;
*/

// new test 
float ffactor = 1;
float gain_kp = 18;
float gain_ki = 0.4;
float gain_kd = 15;

/* - - - - NON-EXTERN GLOBALS AND FUNCTION PROTOTYPES - - - - */
void initTaskTimer(void);
void initLCD(void);
void setupIO(void);
uint8_t PID(void);
void measure(void);

/* - - - - INTERRUPT HANDLER - - - - */
ISR (TIMER0_COMPA_vect)
{      
	if (timer_buzzing)
		TOGGLE_BUZZER_PIN;  //toggle D2 to make 500Hz wave
	if (timer_on)
	/* update timer_ms, timer_s, timer_min as necessary */
	{
		if (timer_ms == 0)
		{
			timer_ms = 999;
			if (timer_s == 0)
			{
				timer_s = 59;
				if (timer_min == 0)
				{
					/* stop and disable timer */
					timer_done = 1;
					timer_on = 0;
					timer_s = 0;
					timer_ms = 999;
					if (timer_buzzing)
						timer_buzzing = 0;
					else if (mode == MODE_DCHEAT2 || mode == MODE_COOK)
					{
						/* if in certain modes, stop cooking when the timer expires */
						sstate = setStepper(sstate,STEPPER_MINPOS);
						mode = MODE_HOME;
						/* initate 3 second buzz */
						timer_buzzing = 1;
						setTimerMinSec(0,3);
					}
				}
				else
					timer_min--;
			}
			else
				timer_s--;
		}
		else
			timer_ms--;
	}
	updateStepperTimer();
	if (waitKeypad) waitKeypad--;
	if (waitMeas) waitMeas--;
	if (waitMain) waitMain--;
} 

/* - - - - FUNCTION IMPLEMENTATIONS - - - - */
void initTaskTimer()
{
	// set task time variables to their starting value		
	//setup 1ms timer and interrupt
	TIMSK0 = _BV(OCIE0A); // set the timer0 outputA output compare (OC) interrupt enable bit.  
	OCR0A = 249; // set the timer0 outputA compare value. (output compare match happens when TCNT0 hits 249)
	TCCR0B = _BV(CS01) | _BV(CS00);  // set timer0 clock prescaler to 64.  16MHz/64 = 250kHz
	TCCR0A = _BV(WGM01); //CTC mode
}

inline void setTimerMinSec(uint8_t min, uint8_t sec)
{
	timer_min = min;
	timer_s = sec;
	timer_ms = 1000;
	timer_done = 0;
	timer_on = 1;
}

void initLCD(void)
{
	LCDinit(); //initialize the display
	LCDcursorOFF();
	LCDclr();  //clear the display
	LCDGotoXY(0,0);
	_delay_ms(50);
}

void softReset(void)
{
	sstate = 0;
	T = 0;
	Td = 0;
}

/*! Assigning the probe selection
-TODO right now it's default IC only
*/
void selectProbe()
{   /*
	float testMeas;
	//NOTE: if Td > 140 C, must select IC probe
	
	
	//pull A1 high with internal pull-up resistor
	PORTA |= 1<<1;
	initIC();
	_delay_ms(5);
	//read analog voltage on A1
	//if it exceeds 4.75V = 973 LSBs, the IC probe is disconnected
	testMeas = readIC();
	PORTA &= ~(1<<1); //disable A1 pullup now
	if (testMeas > 4.75)
	{
		probe = PROBE_NTC;
	}
	else */
		initIC();
		probe = PROBE_IC;
}
/*! Input/Output port definition
*/
void setupIO()
{
	DDRA = 0x00;  // sensor board is all inputs.  readNTC.c will set resister outputs as necessary
	PORTA = 0;    // disable pull-up resistors
	DDRB = 0xFF;  // keypad is outputs
	PORTB = 0;
	DDRC = 0xFF;  // LCD is outputs
	PORTC = 0;
	DDRD = 0xF0;  // stepper motor pins are outputs (high nibble), TX is output
	PORTD = 0;
}

/*! PID control algorithm

*/
uint8_t PID()
{
	int8_t newpos;
	preverror = error;
	error = Td - T;

	
	if (error > 0 && preverror > 0)
		accerror = accerror*ffactor + error;
	else if (error < 0 && preverror < 0)
		accerror = accerror*ffactor + error;
	else
		accerror = 0;

	//fprintf(stdout,"e=%i \t ae=%i \t de=%i\n\r",(int16_t)error,(int16_t)accerror,(int16_t)(error-preverror));
	
	Q = gain_kp*error + gain_ki*accerror + gain_kd*(error-preverror);

	newpos = (int8_t)((float)(stepper_maxpos - STEPPER_MINPOS)*Q/(float)QMAX) + STEPPER_MINPOS; 
	
	if (Td > 20)  //we're actually cooking and not trying to stop
		newpos = newpos + stepper_onpos - 1;  //compensate for the difference between "off" and when hornilla actually turns on
	
	if (newpos > max_sstate)
		newpos = max_sstate;
		
	if (error < Td_deadzone)
		timer_on = 1;  // start the cooking timer (which was previous set, but turned off)
		
	//fprintf(stdout,"newpos=%i\n\r",newpos);
	
	return newpos;// the new position
}

/*! General measuring task function
Retrieves measurement from specific probe measuring functions.
Performes measurement averaging
*/
inline void measure()
{	
	float tempT;
	//fprintf(stdout,"rtc=%u\n\r",readTempCount);
	if (readTempCount < NUM_MEAS)
	{
		//Reading Temperature
		if (probe == PROBE_NTC)
		{
			tempT = readNTC();
			currT += tempT;
			//fprintf(stdout,"currT = %u with NTC\n\r",(uint16_t)(tempT));
		}
		else
		{
			tempT = readIC();
			currT += tempT;
			//fprintf(stdout,"currT = %u with IC\n\r",(uint16_t)(tempT));
		}
		readTempCount++;
		if (readTempCount == NUM_MEAS)
		{
			T = currT/((float)NUM_MEAS);
			//fprintf(stdout,"T = %u\n\r",(uint16_t)(tempT*10.0));
		}
	}	
}

int main()
{
	uint8_t kp;
	cli();
	
	// if necessary, re-init the EEPROM values
	if ((eeprom_read_byte((uint8_t*)EEPROM_BYTE_CHECK0) != EEPROM_CHECKVAL) || (eeprom_read_byte((uint8_t*)EEPROM_BYTE_CHECK1) != EEPROM_CHECKVAL))
	{
		//EEPROM has not be written to yet, so set the record to a default of 1000
		eeprom_write_byte((uint8_t*)EEPROM_BYTE_CHECK0, EEPROM_CHECKVAL);  
		eeprom_write_byte((uint8_t*)EEPROM_BYTE_CHECK1, EEPROM_CHECKVAL);
		eeprom_write_byte((uint8_t*)EEPROM_BYTE_MAXPOS, MAXPOS_DEFAULT);
		eeprom_write_byte((uint8_t*)EEPROM_BYTE_ONPOS, ONPOS_DEFAULT);
		stepper_maxpos = MAXPOS_DEFAULT;
	} 
	else //retrieve previous stepper motor position calibration data
	{
		stepper_maxpos = eeprom_read_byte((uint8_t*)EEPROM_BYTE_MAXPOS);
		stepper_onpos = eeprom_read_byte((uint8_t*)EEPROM_BYTE_ONPOS);
	}
	
	initTaskTimer();
	setupIO();

	mode = MODE_WELCOME;


	
	//setup variables for inputing parameters
	//since 0x00 is the string terminator, there must be a 0x00 in the string -- that's why is char[NUMDIGITS+1]
	//[0x31][0x32][0x30][0x00][0x00] = "120" --> 120
	
	#ifdef DEBUG
		uart_init();
		stdout = stdin = stderr = &uart_str;
		fprintf(stdout,"Hellow PC!\n\r"); 
	#endif
	
	_delay_ms(50);
	initLCD();
	initStepper();
	selectProbe();
	
	//starting task times
	waitStepper = WAIT_STEPPER_MS; 
	waitMain = WAIT_MAIN_MS;
	waitKeypad = WAIT_KEY_MS;
	waitMeas = WAIT_MEAS_MS;
	
	sei();
	updateLCD();

	while(1)
	{
		//fprintf(stdout,"%u,%u,%u\n\r",timer_ms,timer_s,timer_min);
		taskStepper(); // executes an action if waitStepper==0 AND stepper is enabled
		if (waitKeypad == 0)
		{ 
			waitKeypad = WAIT_KEY_MS; 
			kp = scanKeypad();
			if (kp)
			{

				updateState(kp);
			}
		}
		if (waitMain == 0)
		{
			waitMain = WAIT_MAIN_MS;
			if (mainToggle)
			//this half is for 1s interval
			{
				if (mode == MODE_COOK)
					sstate = setStepper(sstate,PID());
				updateLCD();
				mainToggle = 0;
			}
			else
			{
				fprintf(stdout,"m");
				mainToggle = 1;
				// let measure() start over by reset count and accumulator
				currT = 0;
				readTempCount = 0; 
			}
		}
		if ((mainToggle) && (waitMeas==0))
		{
			measure();
			waitMeas = WAIT_MEAS_MS;
		}
	}
}


